@@ -1,3 +1,4 @@
+import os
 import re
 import sys
 import copy
@@ -9,6 +10,9 @@ import psycopg2.extras
 import psycopg2.extensions
 import numpy as np
 import pandas as pd
+import threading
+import csv
+import tempfile
 
 
 class Json:
@@ -142,6 +146,14 @@ class PgDatabase:
         self.connect()
         return False if self._connection is None else True
 
+    def is_in_transaction(self):
+        if self._connection is None:
+            return False
+        if self._connection.status == psycopg2.extensions.STATUS_READY \
+                and self._connection.isolation_level == psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED:
+            return True
+        return self._connection.status == psycopg2.extensions.STATUS_IN_TRANSACTION
+
     def begin(self):
         if not self.connect_to_db():
             return False
@@ -327,6 +339,99 @@ class PgDatabase:
         result_sql = 'DELETE FROM {} WHERE {}'.format(table, act_where_sql)
         return self._execute_sql(sql_string=result_sql, sql_params=act_where_values, get_result=False)
 
+    def execute_sql_without_result(self, result_sql):
+        self._execute_sql(sql_string=result_sql, get_result=False)
+
+    def copy_from(self, table, data, through_pipe=True, csv_filepath=None):
+        err_str = f'Error in {self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
+
+        def copy_from_csv_file(csv_file):
+            try:
+                sql = f"""
+                    COPY {table}({','.join(data[0].keys())}) FROM STDIN WITH
+                    CSV
+                    HEADER
+                    DELIMITER AS ','
+                    """
+                cursor.copy_expert(sql, csv_file)
+                return True
+            except Exception as e:
+                print(f'{err_str}.copy_from_csv_file(): {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
+                self._connection.rollback()
+                return False
+
+        close_connection = True if self._connection is None else False
+        if not self.connect_to_db():
+            return False
+        cursor = self._connection.cursor()
+
+        for index, row in enumerate(data):
+            for key in row:
+                if isinstance(row[key], str):
+                    row[key] = row[key].replace('\t', '\\t').replace('\n', '\\n').\
+                        replace('\r', '\\r').replace('`', "'").replace('\\', '')
+                if row[key] == "":
+                    row[key] = "\\N"
+                if row[key] is None:
+                    row[key] = "\\N"
+                if isinstance(row[key], dict):
+                    row[key] = json.dumps(row[key], ensure_ascii=False)
+
+        if through_pipe:
+            # https://stackoverflow.com/questions/6765310/piping-postgres-copy-in-python-with-psycopg2
+            r_fd, w_fd = os.pipe()
+            to_thread = threading.Thread(target=copy_from_csv_file, args=(os.fdopen(r_fd),))
+            to_thread.start()
+            write_f = os.fdopen(w_fd, 'w')
+            wr = csv.DictWriter(write_f, tuple(data[0].keys()), delimiter=',', lineterminator='\n')
+            wr.writeheader()
+            for row in data:
+                wr.writerow(row)
+            write_f.close()
+            to_thread.join()
+        else:
+            if csv_filepath is None:
+                csv_filepath = os.path.join(tempfile.gettempdir(), '__temp.csv')
+            if os.path.exists(csv_filepath):
+                os.remove(csv_filepath)
+
+            with open(csv_filepath, 'wt', encoding='utf-8') as csv_file:
+                wr = csv.DictWriter(csv_file, tuple(data[0].keys()), delimiter=',', lineterminator='\n')
+                wr.writeheader()
+                for row in data:
+                    wr.writerow(row)
+
+            copy_from_csv_file(open(csv_filepath, 'rt', encoding='utf-8'))
+            if os.path.exists(csv_filepath):
+                os.remove(csv_filepath)
+
+        self._connection.commit()
+        if close_connection:
+            self.disconnect()
+
+    def insert_df(self, table, df):
+        df_columns = list(df.columns.values)
+        columns = ",".join(df_columns)
+        values = "VALUES({})".format(",".join(["%s" for _ in df_columns]))
+        insert_stmt = "INSERT INTO {} ({}) {}".format(table, columns, values)
+        close_connection = True if self._connection is None else False
+        if not self.connect_to_db():
+            return False
+        cursor = self._connection.cursor()
+        psycopg2.extras.execute_batch(cursor, insert_stmt, df.values)
+        self._connection.commit()
+        if close_connection:
+            self.disconnect()
+
+    def create_database(self):
+        sql_string = 'CREATE DATABASE {dbname};'.format(**self._settings)
+        return self._execute_sql(sql_string=sql_string, get_result=False)
+
+    def create_schema(self):
+        sql_string = "CREATE SCHEMA {schema} AUTHORIZATION {user};".format(**self._settings)
+        return self._execute_sql(sql_string=sql_string, get_result=False)
+
+
     def _test_table(self, table_name, table_columns, create, primary_key=None):
         func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
         err_str = f'Error in {func_name}'
@@ -504,6 +609,7 @@ class TasksDatabase:
                          'max_retry_count': 'int4',
                          'scheduled_time': 'timestamp',
                          'defer_time': 'timestamp',
+                         'create_time': 'timestamp',
                          'started_time': 'timestamp',
                          'finished_time': 'timestamp',
                          'priority': 'int4',
@@ -645,7 +751,7 @@ class TasksDatabase:
             select 
                 * 
             from 
-                public.tasks 
+                {schema}.tasks 
             where 
                 status='error'
                 and max_retry_count > retry_count
@@ -663,7 +769,7 @@ class TasksDatabase:
             select 
                 * 
             from 
-                public.tasks 
+                {schema}.tasks 
             where 
                 status is null
                 and scheduled_time is null
