@@ -95,7 +95,7 @@ class Worker(object):
                     if not isinstance(params, dict):
                         params = dict()
                     print(f'{func_name}: try to start "{task_module}.{task_func}"...')
-                    result = func(params)
+                    result = func(**params)
                 except Exception as e:
                     error = f'Error in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}'
                     print(error)
